# AUTOGENERATED! DO NOT EDIT! File to edit: losses.ipynb (unless otherwise specified).

__all__ = ['mse_loss', 'mae_loss', 'binary_cross_entropy', 'ce_loss', 'ce4softmax']

# Cell
import torch
import torch.nn as nn
import torch.nn.functional as F

# Cell
mse_loss = mean_squared_loss = mse = F.mse_loss

# Cell
mae_loss = mean_absolute_error = mae = F.l1_loss

# Cell
binary_cross_entropy = bce_loss = bce = F.binary_cross_entropy

# Cell
ce_loss = F.cross_entropy

# Cell
def ce4softmax(input, target):
    eps = 1e-9
    logp = torch.log(input + eps)
    return ce_loss(logp, target)


# AUTOGENERATED! DO NOT EDIT! File to edit: Model.ipynb (unless otherwise specified).

__all__ = ['Sequential']

# Cell
import torch
import torch.nn as nn

from functools import partial
from fastai import *
from fastai.basic_train import Learner

from .data import create_db
from .layers import *
from .losses import *

from IPython.display import clear_output

# Cell
class Sequential:
    def __init__(self, model=None):
        self.layers = []
        self.last_dim = None
        self.model = model
        self.device = torch.device('cpu')
        if torch.cuda.is_available():
            self.device = torch.device('cuda')

    def add(self, layer):
        layer = layer.get_layer(self.last_dim)
        self.last_dim = layer['output_dim']
        self.layers.extend(layer['layers'])

    def compile(self, loss, optimizer=None):
        if len(self.layers) > 0:
            self.model = nn.Sequential(*self.layers)
        self.loss = loss

    def fit(self, x, y, bs, epochs, lr=1e-3, one_cycle=True, get_lr=True):
        db = create_db(x, y, bs=bs)
        self.learn = Learner(db, self.model, loss_func=self.loss)
        if one_cycle:
            self.learn.fit_one_cycle(epochs, lr)
        else:
            self.learn.fit(epochs, lr)

    def lr_find(self, x, y, bs):
        db = create_db(x, y, bs=bs)
        learn = Learner(db, self.model, loss_func=self.loss)
        learn.lr_find()
        clear_output()
        learn.recorder.plot(suggestion=True)

    def predict(self, x):
        self.learn.model.eval()
        with torch.no_grad():
            y_preds = self.learn.model(torch.Tensor(x).to(device))
        return y_preds.cpu().numpy()


# AUTOGENERATED! DO NOT EDIT! File to edit: Activations.ipynb (unless otherwise specified).

__all__ = ['relu', 'ReLU', 'LeakyReLU', 'leaky_relu', 'Mish', 'mish', 'sigmoid', 'softmax', 'get_activation',
           '__activations__']

# Cell
import torch
import torch.nn as nn
import torch.nn.functional as F

# Cell
def relu(inplace=True):
    return nn.ReLU(inplace=inplace)

ReLU = Relu = relu

# Cell
def LeakyReLU(inplace=True, negative_slope=0.01):
    return nn.LeakyReLU(negative_slope, inplace=inplace)

leaky_relu = leakyrelu = LeakyReLU

# Cell
class Mish(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x * (torch.tanh(F.softplus(x)))

mish = Mish

# Cell
sigmoid = nn.Sigmoid

# Cell
softmax = nn.Softmax(dim=-1)

# Cell
__activations__ = {
    'relu': relu(),
    'leaky_relu': leaky_relu(),
    'mish': mish(),
    'sigmoid': sigmoid(),
    'softmax': softmax
}

def get_activation(activation):
    return __activations__[activation.lower()]


# AUTOGENERATED! DO NOT EDIT! File to edit: Layers.ipynb (unless otherwise specified).

__all__ = ['__inputDimError__', 'Dense', 'Conv2D', 'MaxPool2D', 'Flatten', 'Activation']

# Cell
import numpy as np
import torch.nn as nn
from fastai.vision import *
from fastai import layers

from .activations import *
from functools import partial

# Cell
class __inputDimError__(Exception):
    pass

# Cell
class Dense:
    def __init__(self, units, input_dim=None, activation=None,
                 use_bias=True, kernel_regularizer=None, bias_regularizer=None,
                 activity_regularizer=None):
        """
        Linear layer that takes in `input_dim` and converts it to `units` number of dimensions.
        parameters:
        - units: output dimension.
        - input_dim: input dimension.
        - activation (optional): non-linear activation.
        - use_bias (optional): To include bias layer or not (default: True)
        """
        super().__init__()

        self.input_dim = input_dim
        self.activation = get_activation(activation) if activation else None
        if input_dim:
            self.layer = nn.Linear(input_dim, units, bias=use_bias)
        else:
            self.layer = partial(nn.Linear, out_features=units, bias=use_bias)
        # TODO: implement regularizers

    def get_layer(self, input_dim=None):
        if input_dim is None and self.input_dim is None:
            __inputDimError__("Need to specify number of input dimensions in first layer")
        elif input_dim:
            self.input_dim = input_dim
            self.layer = self.layer(in_features=input_dim)
        # else self.layer is already is assigned

        self.output_dim = self.layer.out_features
        layers = [layer for layer in [self.layer, self.activation] if layer]

        return {'output_dim': self.output_dim, 'layers': layers}

# Cell
class Conv2D:
    def __init__(self, filters:int, kernel_size:int=3, strides:int=1, padding:int=None,
                 activation:str=None, use_bias:bool=True, input_shape:tuple=None):
        """
        Apply convolution on image using kernel filters.
        parameters:
        - filters: number of kernel filters
        - kernel_size: the width of the (square) kernel
        - strides: number of pixels to skip when sliding kernel (default 1)
        - padding: number of pixels to pad incoming image ` defaults to `ks//2`
        - activation: non-linearity
        - use_bias: bias
        - input_shape: incoming image shape of (#Channels, width, height)
        """
        self.input_shape = input_shape
        if input_shape:
            ni = input_shape[0]
            self.layer = conv2d(ni, filters, kernel_size, strides, padding, use_bias)
        else:
            self.layer = partial(conv2d, nf=filters, ks=kernel_size,
                                 stride=strides, padding=padding, bias=use_bias)
        self.activation = get_activation(activation) if activation else None

    def get_layer(self, input_shape=None):
        if input_shape is None and self.input_shape is None:
            __inputDimError__("Need to specify input shape in first layer")
        elif input_shape:
            self.input_shape = input_shape
            ni = self.input_shape[0]
            self.layer = self.layer(ni=ni)
        # else self.input_shape is already is assigned

        dummy_x = torch.zeros(self.input_shape).unsqueeze(0)
        self.output_shape = self.layer(dummy_x).shape[1:]
        layers = [layer for layer in [self.layer, self.activation] if layer]

        return {'output_dim': self.output_shape, 'layers': layers}

# Cell
class MaxPool2D:
    def __init__(self, kernel_size:int=2, strides:int=None, padding:int=0,
                 input_shape:tuple=None):
        """
        Apply convolution on image using kernel filters.
        parameters:
        - filters: number of kernel filters
        - kernel_size: the width of the (square) kernel
        - strides: number of pixels to skip when sliding kernel (default 1)
        - padding: number of pixels to pad incoming image ` defaults to `ks//2`
        - activation: non-linearity
        - use_bias: bias
        - input_shape: incoming image shape of (#Channels, width, height)
        """
        self.input_shape = input_shape
        self.layer = nn.MaxPool2d(kernel_size, strides, padding)


    def get_layer(self, input_shape=None):
        if input_shape is None and self.input_shape is None:
            __inputDimError__("Need to specify input shape in first layer")
        elif input_shape:
            self.input_shape = input_shape
        # else self.input_shape is already is assigned

        dummy_x = torch.zeros(self.input_shape).unsqueeze(0)
        self.output_shape = self.layer(dummy_x).shape[1:]
        layers = [self.layer]

        return {'output_dim': self.output_shape, 'layers': layers}

# Cell
class Flatten:
    def __init__(self, input_shape=None):
        self.layer = layers.Flatten()
        self.input_shape = input_shape

    def get_layer(self, input_shape=None):
        if input_shape is None and self.input_shape is None:
            __inputDimError__("Need to specify input shape in first layer")
        elif input_shape:
            self.input_shape = input_shape
        # else self.input_shape is already is assigned

        self.output_dim = np.prod(self.input_shape)
        layers = [self.layer]

        return {'output_dim': self.output_dim, 'layers': layers}

# Cell
class Activation:
    def __init__(self, activation, input_shape=None):
        self.layer = get_activation(activation)
        self.input_shape = input_shape
        self.output_dim = input_shape

    def get_layer(self, input_shape=None):
        if input_shape is None and self.input_shape is None:
            __inputDimError__("Need to specify input shape in first layer")
        elif input_shape:
            self.input_shape = input_shape
            self.output_dim = input_shape
        # else self.input_shape is already is assigned

        layers = [self.layer]

        return {'output_dim': self.output_dim, 'layers': layers}


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"relu": "Activations.ipynb",
         "ReLU": "Activations.ipynb",
         "LeakyReLU": "Activations.ipynb",
         "leaky_relu": "Activations.ipynb",
         "Mish": "Activations.ipynb",
         "mish": "Activations.ipynb",
         "sigmoid": "Activations.ipynb",
         "softmax": "Activations.ipynb",
         "get_activation": "Activations.ipynb",
         "__activations__": "Activations.ipynb",
         "__inputDimError__": "Layers.ipynb",
         "Dense": "Layers.ipynb",
         "Conv2D": "Layers.ipynb",
         "MaxPool2D": "Layers.ipynb",
         "Flatten": "Layers.ipynb",
         "Activation": "Layers.ipynb",
         "Sequential": "Model.ipynb",
         "Data": "data.ipynb",
         "TestData": "data.ipynb",
         "TrainData": "data.ipynb",
         "create_db": "data.ipynb",
         "mse_loss": "losses.ipynb",
         "mae_loss": "losses.ipynb",
         "binary_cross_entropy": "losses.ipynb",
         "ce_loss": "losses.ipynb",
         "ce4softmax": "losses.ipynb"}

modules = ["activations.py",
           "layers.py",
           "model.py",
           "data.py",
           "losses.py"]

doc_url = "https://sachinruk.github.io/keraTorch/"

git_url = "https://github.com/sachinruk/keraTorch/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: data.ipynb (unless otherwise specified).

__all__ = ['Data', 'TrainData', 'TestData', 'create_db']

# Cell
import torch
from torch.utils.data import Dataset, DataLoader
from fastai.data_block import DataBunch, DatasetType

from sklearn.model_selection import train_test_split

# import warnings

# torch.Tensor.ndim = property(lambda x: x.dim())
# tt = torch.Tensor

# Cell
class Data(Dataset):
    """
    Load raw x,y data
    """
    def __init__(self, *args):
        super().__init__()
        self.data = args

    def __len__(self):
        return len(self.data[0])

    def __getitem__(self, i):
        return {f'arg_{i}': torch.Tensor([x[i]])
                for i, x in enumerate(self.data)}

# Cell
class TrainData(Dataset):
    """
    Load raw x,y data
    """
    def __init__(self, x, y):
        super().__init__()
        self.x, self.y = x, y

    def __len__(self):
        return len(self.x)

    def __getitem__(self, i):
        return torch.tensor(self.x[i]), torch.tensor(self.y[i])


class TestData(Dataset):
    """
    Load raw x,y data
    """
    def __init__(self, x):
        super().__init__()
        self.x = x

    def __len__(self):
        return len(self.x)

    def __getitem__(self, i):
        return torch.tensor(self.x[i])

# Cell
def create_db(x, y, train_size=0.8, bs=96, random_state=42):
    """
    Take dataframe and convert to Fastai databunch
    """
    X_train, X_test, y_train, y_test = train_test_split(x, y, train_size=train_size)

    train_ds = TrainData(X_train, y_train)
    val_ds = TrainData(X_test, y_test)

    bs = min(bs, len(train_ds))
    val_bs = min(bs, len(val_ds))

    train_dl = DataLoader(train_ds, bs)
    val_dl = DataLoader(val_ds, val_bs)

    return DataBunch(train_dl, val_dl)